@@ -9,7 +9,7 @@ use strict;
 use warnings;
 use base qw( IO::Socket );
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 
 use Carp;
 
@@ -25,6 +25,30 @@ use Socket qw(
    SO_REUSEADDR SO_REUSEPORT SO_BROADCAST
 );
 
+my $IPv6_re = do {
+   # translation of RFC 3986 3.2.2 ABNF to re
+   my $IPv4address = do {
+      my $dec_octet = q<(?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])>;
+      qq<$dec_octet(?: \\. $dec_octet){3}>;
+   };
+   my $IPv6address = do {
+      my $h16  = qq<[0-9A-Fa-f]{1,4}>;
+      my $ls32 = qq<(?: $h16 : $h16 | $IPv4address)>;
+      qq<(?:
+                                            (?: $h16 : ){6} $ls32
+         |                               :: (?: $h16 : ){5} $ls32
+         | (?:                   $h16 )? :: (?: $h16 : ){4} $ls32
+         | (?: (?: $h16 : ){0,1} $h16 )? :: (?: $h16 : ){3} $ls32
+         | (?: (?: $h16 : ){0,2} $h16 )? :: (?: $h16 : ){2} $ls32
+         | (?: (?: $h16 : ){0,3} $h16 )? ::     $h16 :      $ls32
+         | (?: (?: $h16 : ){0,4} $h16 )? ::                 $ls32
+         | (?: (?: $h16 : ){0,5} $h16 )? ::                 $h16
+         | (?: (?: $h16 : ){0,6} $h16 )? ::
+      )>
+   };
+   qr<$IPv6address>xo;
+};
+
 =head1 NAME
 
 C<IO::Socket::IP> - Use IPv4 and IPv6 sockets in a protocol-independent way
@@ -126,26 +150,79 @@ If the constructor fails, it will set C<$@> to an appropriate error message;
 this may be from C<$!> or it may be some other string; not every failure
 necessarily has an associated C<errno> value.
 
+If either C<LocalHost> or C<PeerHost> (or their C<...Addr> synonyms) have any
+of the following special forms, they are split to imply both the hostname and
+service name:
+
+ hostname.example.org:port    # DNS name
+ 10.0.0.1:port                # IPv4 address
+ [fe80::123]:port             # IPv6 address
+
+In each case, C<port> is passed to the C<LocalService> or C<PeerService>
+argument.
+
+Either of C<LocalService> or C<PeerService> (or their C<...Port> synonyms) can
+be either a service name, a decimal number, or a string containing both a
+service name and number, in the form
+
+ name(number)
+
+In this case, the name will be tried first, but if the resolver does not
+understand it then the port number will be used instead.
+
+=head1 $sock = IO::Socket::IP->new( $peeraddr )
+
+As a special case, if the constructor is passed a single argument (as
+opposed to an even-sized list of key/value pairs), it is taken to be the value
+of the C<PeerAddr> parameter. The example in the SYNOPSIS section may also
+therefore be written as
+
+ my $sock = IO::Socket::IP->new( "www.google.com:www" )
+    or die "Cannot construct socket - $@";
+
 =cut
 
+sub new 
+{
+   my $class = shift;
+   my %arg = (@_ == 1) ? (PeerHost => $_[0]) : @_;
+
+   $arg{PeerHost} = delete $arg{PeerAddr}
+      if exists $arg{PeerAddr} && !exists $arg{PeerHost};
+
+   $arg{PeerService} = delete $arg{PeerPort}
+      if exists $arg{PeerPort} && !exists $arg{PeerService};
+
+   $arg{LocalHost} = delete $arg{LocalAddr}
+      if exists $arg{LocalAddr} && !exists $arg{LocalHost};
+
+   $arg{LocalService} = delete $arg{LocalPort}
+      if exists $arg{LocalPort} && !exists $arg{LocalService};
+
+   for my $type (qw(Peer Local)) {
+      my $host    = $type . 'Host';
+      my $service = $type . 'Service';
+
+      if (exists $arg{$host} && !exists $arg{$service}) {
+         local $_ = $arg{$host};
+         defined or next;
+         if (/\A\[($IPv6_re)\](?::([^\s:]*))?\z/o || /\A([^\s:]*):([^\s:]*)\z/) {
+            $arg{$host}    = $1;
+            $arg{$service} = $2 if defined $2 && length $2;
+         }
+      }
+   }
+   return $class->SUPER::new(%arg);
+}
+
 sub configure
 {
    my $self = shift;
    my ( $arg ) = @_;
 
-   # Some conveniences for IO::Socket::INET users
-   for (qw( Local Peer )) {
-      # TODO: Consider hostname:port notations
-      !defined $arg->{$_."Host"} and defined $arg->{$_."Addr"} and
-         $arg->{$_."Host"} = delete $arg->{$_."Addr"};
-
-      !defined $arg->{$_."Service"} and defined $arg->{$_."Port"} and
-         $arg->{$_."Service"} = delete $arg->{$_."Port"};
-   }
-
    my %hints;
-   my @sockaddrs;
-   my @peeraddrs;
+   my @localinfos;
+   my @peerinfos;
 
    my @sockopts_enabled;
 
@@ -171,7 +248,14 @@ sub configure
       my $host = delete $arg->{LocalHost};
       my $service = delete $arg->{LocalService};
 
-      ( my $err, @sockaddrs ) = getaddrinfo( $host, $service, \%hints );
+      defined $service and $service =~ s/\((\d+)\)$// and
+         my $fallback_port = $1;
+
+      ( my $err, @localinfos ) = getaddrinfo( $host, $service, \%hints );
+
+      if( $err and defined $fallback_port ) {
+         ( $err, @localinfos ) = getaddrinfo( $host, $fallback_port, \%hints );
+      }
 
       $err and ( $@ = "$err", return );
    }
@@ -182,7 +266,14 @@ sub configure
       defined( my $service = delete $arg->{PeerService} ) or
          croak "Expected 'PeerService'";
 
-      ( my $err, @peeraddrs ) = getaddrinfo( $host, $service, \%hints );
+      defined $service and $service =~ s/\((\d+)\)$// and
+         my $fallback_port = $1;
+
+      ( my $err, @peerinfos ) = getaddrinfo( $host, $service, \%hints );
+
+      if( $err and defined $fallback_port ) {
+         ( $err, @peerinfos ) = getaddrinfo( $host, $fallback_port, \%hints );
+      }
 
       $err and ( $@ = "$err", return );
    }
@@ -193,7 +284,7 @@ sub configure
 
    my $listenqueue = delete $arg->{Listen};
 
-   croak "Cannot Listen with a PeerHost" if defined $listenqueue and @peeraddrs;
+   croak "Cannot Listen with a PeerHost" if defined $listenqueue and @peerinfos;
 
    keys %$arg and croak "Unexpected keys - " . join( ", ", sort keys %$arg );
 
@@ -201,41 +292,38 @@ sub configure
    my $binderr;
    my $connecterr;
 
-   if( @peeraddrs ) {
-      # Loop on socket / bind? / connect
-      # TODO: Local binds
+   foreach my $local ( @localinfos ? @localinfos : {} ) {
+      foreach my $peer ( @peerinfos ? @peerinfos : {} ) {
 
-      while( @peeraddrs ) {
-         my $addr = shift @peeraddrs;
+         next if defined $local->{family}   and defined $peer->{family}   and
+            $local->{family} != $peer->{family};
+         next if defined $local->{socktype} and defined $peer->{socktype} and
+            $local->{socktype} != $peer->{socktype};
+         next if defined $local->{protocol} and defined $peer->{protocol} and
+            $local->{protocol} != $peer->{protocol};
 
-         $self->socket( @{$addr}{qw( family socktype protocol )} ) or ( $socketerr = $!, next );
+         my $family   = $local->{family}   || $peer->{family}   or next;
+         my $socktype = $local->{socktype} || $peer->{socktype} or next;
+         my $protocol = $local->{protocol} || $peer->{protocol};
+
+         $self->socket( $family, $socktype, $protocol ) or ( $socketerr = $!, next );
 
          foreach my $sockopt ( @sockopts_enabled ) {
             $self->setsockopt( SOL_SOCKET, $sockopt, pack "i", 1 ) or ( $@ = "$!", return );
          }
 
-         $self->connect( $addr->{addr} ) or ( $connecterr = $!, next );
-
-         return $self;
-      }
-   }
-   else {
-      # Loop on socket / bind
-      while( @sockaddrs ) {
-         my $addr = shift @sockaddrs;
-
-         $self->socket( @{$addr}{qw( family socktype protocol )} ) or ( $socketerr = $!, next );
-
-         foreach my $sockopt ( @sockopts_enabled ) {
-            $self->setsockopt( SOL_SOCKET, $sockopt, pack "i", 1 ) or ( $@ = "$!", return );
+         if( defined( my $addr = $local->{addr} ) ) {
+            $self->bind( $addr ) or ( $binderr = $!, next );
          }
 
-         $self->bind( $addr->{addr} ) or ( $binderr = $!, next );
-
          if( defined $listenqueue ) {
             $self->listen( $listenqueue ) or ( $@ = "$!", return );
          }
 
+         if( defined( my $addr = $peer->{addr} ) ) {
+            $self->connect( $addr ) or ( $connecterr = $!, next );
+         }
+
          return $self;
       }
    }
@@ -376,22 +464,17 @@ __END__
 
 =item *
 
-Accept C<hostname:port> notation in C<LocalAddr> and C<PeerAddr> constructor
-args, as per C<IO::Socket::INET>. Special care should be taken on raw IPv6
-addresses, which need wrapping as C<[2001::0123:4567]:89>.
-
-=item *
-
-Support C<LocalHost> and C<LocalService> to C<bind()> even when also
-C<connect()>ing to a peer.
-
-=item *
-
 Cache the returns from C<sockhost_service> and C<peerhost_service> to avoid
 double-lookup overhead in such code as
 
   printf "Peer is %s:%d\n", $sock->peerhost, $sock->peerport;
 
+=item *
+
+Implement constructor args C<Timeout>, C<Blocking> and maybe C<Domain>. Except
+that C<Domain> is harder because L<IO::Socket> wants to dispatch to subclasses
+based on it. Maybe C<Family> might be a better name?
+
 =back
 
 =head1 AUTHOR
